"""Inspection Viewer app behavior."""
from . import ObjectClass, TypeHorWeld
from .csvfile import row, File as FileBase


class TypeMarker:
    """Types of marker."""

    VALVE = 0
    MARKER = 1
    MAGNET = 2
    OTVOD = 3
    TROYNIK = 4
    CASE_START = 5
    CASE_END = 6
    REPAIR = 7
    LOAD = 8
    TURN_START = 9
    TURN_END = 10
    FEATURE = 11
    CURVE_SECTION = 12
    TURN_SEGMENT = 13


class TypeDefekt:
    """Types of defekt."""

    CORROZ = 0
    MECHANIC = 1
    DENT = 2
    DENT_METAL_LOSS = 3
    GOFRA = 4
    GWAN = 5
    TECHNOLOGY = 6
    FACTORY = 7
    ADDITIONAL_METAL = 8
    OTHER = 9
    CRACKS_HOR = 10
    CRACK_LIKE = 11
    CRACK_WELD = 12
    LAMINATION = 13
    ANOMALY_HOR_WELD = 14
    ANOMALY_SPIRAL_WELD = 15
    ELLIPSE = 16
    PODZHIG = 17
    GRINDING = 18


MLOSS = [
  TypeDefekt.CORROZ,
  TypeDefekt.GRINDING,
]

DENTS = [
  TypeDefekt.DENT,
  TypeDefekt.DENT_METAL_LOSS,
]

AT_WELD = [
  TypeDefekt.GWAN,
  TypeDefekt.CRACK_WELD,
]

AT_SEAM = [
  TypeDefekt.ANOMALY_HOR_WELD,
  TypeDefekt.ANOMALY_SPIRAL_WELD,
]

VALVE = [
  TypeMarker.VALVE,
]

DEFEKTS = {
  TypeDefekt.CORROZ: "Коррозия",
  TypeDefekt.MECHANIC: "Мех. повреждение",
  TypeDefekt.DENT: "Вмятина",
  TypeDefekt.DENT_METAL_LOSS: "Вмятина с дефектами потери металла",
  TypeDefekt.GOFRA: "Гофра",
  TypeDefekt.GWAN: "Аномалия кольцевого шва",
  TypeDefekt.TECHNOLOGY: "Технологический дефект",
  TypeDefekt.FACTORY: "Заводской дефект",
  TypeDefekt.ADDITIONAL_METAL: "Дополнительный металл/материал",
  TypeDefekt.OTHER: "Другое",
  TypeDefekt.CRACKS_HOR: "Зона продольных трещин",
  TypeDefekt.CRACK_LIKE: "Трещиноподобный дефект",
  TypeDefekt.CRACK_WELD: "Трещина на кольцевом шве",
  TypeDefekt.LAMINATION: "Расслоение",
  TypeDefekt.ANOMALY_HOR_WELD: "Аномалия продольного шва",
  TypeDefekt.ANOMALY_SPIRAL_WELD: "Аномалия спирального шва",
  TypeDefekt.ELLIPSE: "Эллипсность",
  TypeDefekt.PODZHIG: "Поджиг",
  TypeDefekt.GRINDING: "Зашлифовка",
}

LINEOBJ = {
  TypeMarker.VALVE: "Кран",
  TypeMarker.MARKER: "Маркер",
  TypeMarker.MAGNET: "Маркер магнитный",
  TypeMarker.OTVOD: "Отвод-врезка",
  TypeMarker.TROYNIK: "Тройник",
  TypeMarker.CASE_START: "Патрон начало",
  TypeMarker.CASE_END: "Патрон конец",
  TypeMarker.REPAIR: "Место ремонта",
  TypeMarker.LOAD: "Пригруз",
  TypeMarker.TURN_START: "Отвод (поворот) начало",
  TypeMarker.TURN_END: "Отвод (поворот) конец",
  TypeMarker.FEATURE: "Особенность",
  TypeMarker.CURVE_SECTION: "Гнутая секция",
  TypeMarker.TURN_SEGMENT: "Сегмент поворота",
}

DEFAULT_MARKERS = [
  TypeMarker.VALVE,
  TypeMarker.MARKER,
  TypeMarker.MAGNET,
]

REVERSE_MARKER = {
  TypeMarker.CASE_START: TypeMarker.CASE_END,
  TypeMarker.CASE_END: TypeMarker.CASE_START,
  TypeMarker.TURN_START: TypeMarker.TURN_END,
  TypeMarker.TURN_END: TypeMarker.TURN_START,
}

REVERSE_COMMENTS = {
  'лево': 'право',
  'право': 'лево',
  'начало': 'конец',
  'конец': 'начало',
}

COMMON = {
  ObjectClass.WELD: "Шов",
  ObjectClass.THICK: "Изменение толщины стенки трубы",
  ObjectClass.PIPELINE_CATEGORY: "Категория трубопровода",
  ObjectClass.DIAM: "Переход диаметра",
}

SEAM = {
  TypeHorWeld.HORIZONTAL: "Продольный шов",
  TypeHorWeld.SECOND: "Двойной прод. шов",
  TypeHorWeld.NO_WELD: "Цельнотянутая труба",
  TypeHorWeld.SPIRAL: "Спиральный шов",
  TypeHorWeld.UNKNOWN: "Не определено",
}


class Row(row.Row):
    """Row of Deftable.csv file."""

    @staticmethod
    def get_bool(val):
        """Bool value for IV csv."""
        return 'ИСТИНА' if val else 'ЛОЖЬ'

    @staticmethod
    def name_seam(code):
        """Return text for seam object_code_t field."""
        return SEAM[code]

    @staticmethod
    def name_object(code):
        """Return text for object_code_t field."""
        return COMMON[code]

    @staticmethod
    def defekts_dict():
        """Return dict of available defekts types."""
        return DEFEKTS

    @staticmethod
    def lineobj_dict():
        """Return dict of available lineobject types."""
        return LINEOBJ

    @staticmethod
    def markers_default():
        """Return list of lineobject types that use as markers by default."""
        return DEFAULT_MARKERS

    @staticmethod
    def markers_reverse():
        """Return dict of markers for reverse."""
        return REVERSE_MARKER

    @staticmethod
    def comment_reverse():
        """Return dict of comment substrings for reverse."""
        return REVERSE_COMMENTS

    @staticmethod
    def mloss_dict():
        """Return dict of available metal loss defects."""
        return MLOSS

    @staticmethod
    def dents_dict():
        """Return dict of available dent defects."""
        return DENTS

    @staticmethod
    def atweld_dict():
        """Return dict of available at weld defects."""
        return AT_WELD

    @staticmethod
    def atseam_dict():
        """Return dict of available at seam defects."""
        return AT_SEAM

    @staticmethod
    def valve_dict():
        """Return dict of available valve like obects."""
        return VALVE


class File(FileBase):
    """Export/import Deftable.csv file."""

    RowCls = Row
    ENCODING = 'windows-1251'



"""Row with type Defect."""
from pipeline_csv.orientation import Orientation


class Anomaly:
    """Anomaly at the pipe."""

    def __init__(self, row, pipe):
        """Create anomaly at the pipe from csv Row."""
        self.row = row
        self.pipe = pipe

        self.orient1 = None
        if self.row.orient_td:
            self.orient1 = Orientation.from_csv(self.row.orient_td)

        self.orient2 = None
        if self.row.orient_bd:
            self.orient2 = Orientation.from_csv(self.row.orient_bd)

    def __str__(self):
        """As text."""
        return '{} at {}'.format(self.row.object_code_t, self.pipe)

    @property
    def length(self):
        """Return object length as integer."""
        if self.row.length:
            return int(self.row.length)
        return 0

    @property
    def code(self):
        """Return object code as integer."""
        return int(self.row.object_code)

    @property
    def to_left_weld(self):
        """Return distance (mm) from left defect border to upstream weld."""
        return self.row.dist - self.pipe.dist

    @property
    def to_right_weld(self):
        """Return distance (mm) from right defect border to downstream weld."""
        return self.pipe.dist + self.pipe.length - self.row.dist - self.length


class Defect(Anomaly):
    """Defect at the pipe."""

    def __init__(self, row, pipe):
        """Create defect at the pipe from csv Row."""
        super().__init__(row, pipe)

        self.orient_mp = None
        if self.row.mpoint_orient:
            self.orient_mp = Orientation.from_csv(self.row.mpoint_orient)

    @property
    def number_at_pipe(self):
        """Return object number at pipe as integer."""
        return self.pipe.defects.index(self) + 1

    @property
    def is_metal_loss(self):
        """Return True if metal loss defect."""
        return self.code in self.row.mloss_dict()

    @property
    def is_dent(self):
        """Return True if dent defect."""
        return self.code in self.row.dents_dict()

    @property
    def is_at_weld(self):
        """Return True if weld placed defect."""
        return self.code in self.row.atweld_dict()

    @property
    def is_at_seam(self):
        """Return True if seam placed defect."""
        return self.code in self.row.atseam_dict()

    def _with_mp(prop):  # pylint: disable=no-self-argument
        """Return decorator for property.

        https://stackoverflow.com/questions/1263451/python-decorators-in-classes
        """
        def wrapper(self):
            """Return None if defect does not have maximum depth point, or defect is located on a seam/weld."""
            if self.is_at_weld or self.is_at_seam or (not self.row.mpoint_dist):
                return None
            return prop(self)  # pylint: disable=not-callable

        return wrapper

    @property
    @_with_mp
    def mp_left_weld(self):
        """Return distance (mm) from maximum depth point to upstream weld."""
        return int(self.row.mpoint_dist) - self.pipe.dist

    @property
    @_with_mp
    def mp_right_weld(self):
        """Return distance (mm) from maximum depth point to downstream weld."""
        return self.pipe.dist + self.pipe.length - int(self.row.mpoint_dist)

    @property
    @_with_mp
    def mp_seam(self):
        """Return distance (angle minutes) from maximum depth point to nearest seam."""
        if not self.pipe.to_seam_data:
            return None

        mpoint = Orientation.from_csv(self.row.mpoint_orient)
        dist = mpoint.dist_to(self.pipe.seam1)
        if self.pipe.seam2:
            dist = min(dist, mpoint.dist_to(self.pipe.seam2))

        return self.pipe.minutes2mm(dist)

    @property
    @_with_mp
    def mp_seam_weld(self):
        """Return distance (mm) from maximum depth point to nearest seam/weld.

        If the defect does not have maximum depth point, return None.
        """
        values = [i for i in [self.mp_seam, self.mp_left_weld, self.mp_right_weld] if i is not None]
        return min(values)

    def get_to_seams(self, orient):
        """Return distance (mm) from given orientation to pipe seams."""
        to_seam1, to_seam2 = None, None
        if orient:
            to_seam1 = orient.dist_to(self.pipe.seam1)
            if self.pipe.seam2:
                to_seam2 = orient.dist_to(self.pipe.seam2)

        return (to_seam1, to_seam2)

    @property
    def to_seam(self):
        """Return distance (mm) from defect borders to nearest seam or None if pipe does not have seams."""
        if not self.pipe.to_seam_data:
            return None

        if self.orient1 and self.orient2:
            if self.pipe.seam1.is_inside(self.orient1, self.orient2):
                return 0
            if self.pipe.seam2:
                if self.pipe.seam2.is_inside(self.orient1, self.orient2):
                    return 0

        up_seam1, up_seam2 = self.get_to_seams(self.orient1)
        dn_seam1, dn_seam2 = self.get_to_seams(self.orient2)
        dists = [i for i in [up_seam1, up_seam2, dn_seam1, dn_seam2] if i is not None]
        if not dists:
            return None

        return min(dists)

    @property
    def to_seam_weld(self):
        """Return distance (mm) from defect borders to nearest seam/weld.

        None if pipe does not have seams.
        """
        values = [i for i in [self.to_seam, self.to_left_weld, self.to_right_weld] if i is not None]
        return min(values)

    @property
    def orientation_point(self):
        """Return orientation for defect as point.

        maximum depth point orientation if mp is present
        appropriate orientation if only one from orient1/orient2 is present
        middle point orientation if both orient1 and orient2 is present
        None if no maximum depth point and both orient1 and orient2 is None
        """
        if self.row.mpoint_orient:
            return Orientation.from_csv(self.row.mpoint_orient)

        ornts = [i for i in [self.orient1, self.orient2] if i is not None]

        if len(ornts) == 1:
            return ornts[0]
        if len(ornts) == 2:
            clockwise, _ = self.orient1.dist_to_int(self.orient2)
            ornt = Orientation.from_minutes(self.orient1.as_minutes)
            ornt.add_minutes(int(clockwise / 2))
            return ornt

        return None



"""Interfaces for CSV data stuff."""


def attr_list(cls):
    """Return list of attributes values for class."""
    return [val for key, val in vars(cls).items() if not key.startswith('_')]


class Error(Exception):
    """Package exception."""


class ObjectClass:
    """Classes of csv objects."""

    WELD = 0
    MARKER = 1
    DEFEKT = 2
    THICK = 3
    HOR_WELD = 4
    PIPELINE_CATEGORY = 5
    DIAM = 6


class TypeHorWeld:
    """Types of horizontal weld."""

    HORIZONTAL = 0
    SECOND = 1
    NO_WELD = 2
    SPIRAL = 3
    UNKNOWN = 4


class DefektSide:
    """Types of defekt location."""

    UNKNOWN = 0
    OUTSIDE = 1
    INSIDE = 2
    IN_WALL = 3
